// **********************
// * POD-bot 2.5 bugfix *
// **********************
// version 3.9
// by Pierre-Marie Baty <pm@bots-united.com>
// and AustinBotServer <austinbotserver@mindspring.com>


#include <extdll.h>
#include <dllapi.h>
#include <h_export.h>
#include <meta_api.h>


// type of packets in suspicious messages
#define PACKET_BYTE 1
#define PACKET_CHAR 2
#define PACKET_SHORT 3
#define PACKET_LONG 4
#define PACKET_ANGLE 5
#define PACKET_COORD 6
#define PACKET_STRING 7
#define PACKET_ENTITY 8

// Counter-Strike team IDs
#define TEAM_TERRORIST 1
#define TEAM_COUNTER_TERRORIST 2

// normal / zoomed / scoped FOV widths
#define FOV_NORMAL 90
#define FOV_ZOOMED 55
#define FOV_SCOPED_SHORT 40
#define FOV_SCOPED_FAR 15 // AWP is the exception, it can scope down to 10 degrees


typedef struct
{
   int msg_dest; // first parameter passed in a MESSAGE_BEGIN
   int msg_type; // second parameter passed in a MESSAGE_BEGIN
   const float *pOrigin; // third parameter passed in a MESSAGE_BEGIN
   edict_t *ed; // fourth parameter passed in a MESSAGE_BEGIN
} message_header_t;


typedef struct
{
   int packet_type; // type of packet in a suspicious message
   int iValuePassed; // value passed in the packet, in case it is an integer
   float flValuePassed; // value passed in the packet, in case it is a float
   const char *szValuePassed; // value passed in the packet, in case it is a string
} message_t;


enginefuncs_t g_engfuncs;
globalvars_t  *gpGlobals;
DLL_FUNCTIONS gFunctionTable_Post;
cvar_t botuseradio = {"botuseradio", "1", FCVAR_EXTDLL};
cvar_t botbuygrenades = {"botbuygrenades", "1", FCVAR_EXTDLL};
cvar_t botusescope = {"botusescope", "1", FCVAR_EXTDLL};
bool use_radio = TRUE;
bool buy_grenades = TRUE;
int use_scope = 1;
bool new_suspicious_message = FALSE;
message_header_t suspicious_message_header;
message_t suspicious_message[256]; // up to 256 packets in the message buffer
int suspicious_message_size = 0;
float f_grenadebuy_time = 0;
float f_roundstart_time = 0;

void WINAPI GiveFnptrsToDll (enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals);
int Spawn_Post (edict_t *pent);
void StartFrame_Post (void);
void RunPlayerMove (edict_t *fakeclient, const float *viewangles, float forwardmove, float sidemove, float upmove, unsigned short buttons, byte impulse, byte msec);
void TraceLine (const float *v1, const float *v2, int fNoMonsters, edict_t *pentToSkip, TraceResult *ptr);
void MessageBegin (int msg_dest, int msg_type, const float *pOrigin, edict_t *ed);
void MessageEnd (void);
void MessageEnd_Post (void);
void WriteByte (int iValue);
void WriteChar (int iValue);
void WriteShort (int iValue);
void WriteLong (int iValue);
void WriteAngle (float flValue);
void WriteCoord (float flValue);
void WriteString (const char *sz);
void WriteEntity (int iValue);
void ExaminateSuspiciousMessage (void);
Vector FakeclientSeesEnemyPlayer (edict_t *pFakeclient);
int GetTeam (edict_t *pEntity);
Vector UTIL_VecToAngles (const Vector &vec);
float UTIL_AngleOfVectors (Vector v1, Vector v2);
float UTIL_WrapAngle (float angle_to_wrap);
void GetGameDir (char *game_dir);


// START of Metamod stuff

enginefuncs_t meta_engfuncs;
enginefuncs_t meta_engfuncs_Post;
gamedll_funcs_t *gpGamedllFuncs;
mutil_funcs_t *gpMetaUtilFuncs;
meta_globals_t *gpMetaGlobals;

META_FUNCTIONS gMetaFunctionTable =
{
   NULL, // pfnGetEntityAPI()
   NULL, // pfnGetEntityAPI_Post()
   NULL, // pfnGetEntityAPI2()
   GetEntityAPI2_Post, // pfnGetEntityAPI2_Post()
   NULL, // pfnGetNewDLLFunctions()
   NULL, // pfnGetNewDLLFunctions_Post()
   GetEngineFunctions, // pfnGetEngineFunctions()
   GetEngineFunctions_Post, // pfnGetEngineFunctions_Post()
};

plugin_info_t Plugin_info = {
   META_INTERFACE_VERSION, // interface version
   "PodFix", // plugin name
   "4.0", // plugin version
   "13/02/2004", // date of creation
   "Pierre-Marie Baty <pm@bots-united.com> & Austin <austinbotserver@mindspring.com>", // plugin author
   "http://racc.bots-united.com", // plugin URL
   "PODFIX", // plugin logtag
   PT_CHANGELEVEL, // when loadable
   PT_ANYTIME, // when unloadable
};


C_DLLEXPORT int Meta_Query (char *ifvers, plugin_info_t **pPlugInfo, mutil_funcs_t *pMetaUtilFuncs)
{
   // this function is the first function ever called by metamod in the plugin DLL. Its purpose
   // is for metamod to retrieve basic information about the plugin, such as its meta-interface
   // version, for ensuring compatibility with the current version of the running metamod.

   // keep track of the pointers to metamod function tables metamod gives us
   gpMetaUtilFuncs = pMetaUtilFuncs;
   *pPlugInfo = &Plugin_info;

   // check for interface version compatibility
   if (strcmp (ifvers, Plugin_info.ifvers) != 0)
   {
      int mmajor = 0, mminor = 0, pmajor = 0, pminor = 0;

      LOG_CONSOLE (PLID, "%s: meta-interface version mismatch (metamod: %s, %s: %s)", Plugin_info.name, ifvers, Plugin_info.name, Plugin_info.ifvers);
      LOG_MESSAGE (PLID, "%s: meta-interface version mismatch (metamod: %s, %s: %s)", Plugin_info.name, ifvers, Plugin_info.name, Plugin_info.ifvers);

      // if plugin has later interface version, it's incompatible (update metamod)
      sscanf (ifvers, "%d:%d", &mmajor, &mminor);
      sscanf (META_INTERFACE_VERSION, "%d:%d", &pmajor, &pminor);

      if ((pmajor > mmajor) || ((pmajor == mmajor) && (pminor > mminor)))
      {
         LOG_CONSOLE (PLID, "metamod version is too old for this plugin; update metamod");
         LOG_ERROR (PLID, "metamod version is too old for this plugin; update metamod");
         return (FALSE);
      }

      // if plugin has older major interface version, it's incompatible (update plugin)
      else if (pmajor < mmajor)
      {
         LOG_CONSOLE (PLID, "metamod version is incompatible with this plugin; please find a newer version of this plugin");
         LOG_ERROR (PLID, "metamod version is incompatible with this plugin; please find a newer version of this plugin");
         return (FALSE);
      }
   }

   return (TRUE); // tell metamod this plugin looks safe
}


C_DLLEXPORT int Meta_Attach (PLUG_LOADTIME now, META_FUNCTIONS *pFunctionTable, meta_globals_t *pMGlobals, gamedll_funcs_t *pGamedllFuncs)
{
   // this function is called when metamod attempts to load the plugin. Since it's the place
   // where we can tell if the plugin will be allowed to run or not, we wait until here to make
   // our initialization stuff, like registering CVARs and dedicated server commands.

   char dirname[256];

   // are we allowed to load this plugin now ?
   if (now > Plugin_info.loadable)
   {
      LOG_CONSOLE (PLID, "%s: plugin NOT attaching (can't load plugin right now)", Plugin_info.name);
      LOG_ERROR (PLID, "%s: plugin NOT attaching (can't load plugin right now)", Plugin_info.name);
      return (FALSE); // returning FALSE prevents metamod from attaching this plugin
   }

   // keep track of the pointers to engine function tables metamod gives us
   gpMetaGlobals = pMGlobals;
   memcpy (pFunctionTable, &gMetaFunctionTable, sizeof (META_FUNCTIONS));
   gpGamedllFuncs = pGamedllFuncs;

   // get the MOD directory name (full path)
   GetGameDir (dirname);

   // is it NOT the Counter-Strike directory ?
   if (strcmp (dirname, "cstrike") != 0)
   {
      // then don't allow this plugin to run
      LOG_CONSOLE (PLID, "%s: plugin NOT attaching (bad dirname: '%s')", Plugin_info.name, dirname);
      LOG_ERROR (PLID, "%s: plugin NOT attaching (bad dirname: '%s')", Plugin_info.name, dirname);
      return (FALSE); // returning FALSE prevents metamod from attaching this plugin
   }

   // print a message to notify about plugin attaching
   LOG_CONSOLE (PLID, "%s: plugin attaching", Plugin_info.name);
   LOG_MESSAGE (PLID, "%s: plugin attaching", Plugin_info.name);

   // ask the engine to register the CVARs this plugin uses
   CVAR_REGISTER (&botuseradio);
   CVAR_REGISTER (&botbuygrenades);
   CVAR_REGISTER (&botusescope);

   return (TRUE); // returning TRUE enables metamod to attach this plugin
}


C_DLLEXPORT int Meta_Detach (PLUG_LOADTIME now, PL_UNLOAD_REASON reason)
{
   // this function is called when metamod unloads the plugin. A basic check is made in order
   // to prevent unloading the plugin if its processing should not be interrupted.

   // is metamod allowed to unload the plugin ?
   if ((now > Plugin_info.unloadable) && (reason != PNL_CMD_FORCED))
   {
      LOG_CONSOLE (PLID, "%s: plugin NOT detaching (can't unload plugin right now)", Plugin_info.name);
      LOG_ERROR (PLID, "%s: plugin NOT detaching (can't unload plugin right now)", Plugin_info.name);
      return (FALSE); // returning FALSE prevents metamod from unloading this plugin
   }

   return (TRUE); // returning TRUE enables metamod to unload this plugin
}


C_DLLEXPORT int GetEntityAPI2_Post (DLL_FUNCTIONS *pFunctionTable, int *interfaceVersion)
{
   gFunctionTable_Post.pfnSpawn	= Spawn_Post;
   gFunctionTable_Post.pfnStartFrame = StartFrame_Post;

   memcpy (pFunctionTable, &gFunctionTable_Post, sizeof (DLL_FUNCTIONS));
   return (TRUE);
}


C_DLLEXPORT int GetEngineFunctions (enginefuncs_t *pengfuncsFromEngine, int *interfaceVersion)
{
   meta_engfuncs.pfnRunPlayerMove = RunPlayerMove;
   meta_engfuncs.pfnTraceLine = TraceLine;
   meta_engfuncs.pfnMessageBegin = MessageBegin;
   meta_engfuncs.pfnMessageEnd = MessageEnd;
   meta_engfuncs.pfnWriteByte = WriteByte;
   meta_engfuncs.pfnWriteChar = WriteChar;
   meta_engfuncs.pfnWriteShort = WriteShort;
   meta_engfuncs.pfnWriteLong = WriteLong;
   meta_engfuncs.pfnWriteAngle = WriteAngle;
   meta_engfuncs.pfnWriteCoord = WriteCoord;
   meta_engfuncs.pfnWriteString = WriteString;
   meta_engfuncs.pfnWriteEntity = WriteEntity;

   memcpy (pengfuncsFromEngine, &meta_engfuncs, sizeof (enginefuncs_t));
   return TRUE;
}


C_DLLEXPORT int GetEngineFunctions_Post (enginefuncs_t *pengfuncsFromEngine, int *interfaceVersion)
{
   meta_engfuncs_Post.pfnMessageEnd = MessageEnd_Post;

   memcpy (pengfuncsFromEngine, &meta_engfuncs_Post, sizeof (enginefuncs_t));
   return TRUE;
}

// END of Metamod stuff




#ifdef _WIN32
BOOL WINAPI DllMain (HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
   return TRUE; // required Win32 DLL entry point
}
#endif


void WINAPI GiveFnptrsToDll (enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals)
{
   memcpy (&g_engfuncs, pengfuncsFromEngine, sizeof (enginefuncs_t));
   gpGlobals = pGlobals;
}


int Spawn_Post (edict_t *pent)
{
   // are we spawning a hostage ?
   if (strcmp (STRING (pent->v.classname), "hostage_entity") == 0)
   {
      pent->v.health = 0x0FFFFFFF; // max out the hostage life points
      pent->v.max_health = 0x0FFFFFFF;
   }

   RETURN_META_VALUE (MRES_IGNORED, 0);
}


void StartFrame_Post (void)
{
   int client_index;
   edict_t *pClient;

   // retrieve our CVARs...
   use_radio = (CVAR_GET_FLOAT ("botuseradio") > 0);
   buy_grenades = (CVAR_GET_FLOAT ("botbuygrenades") > 0);
   use_scope = CVAR_GET_FLOAT ("botusescope");

   // is it time to make the bots buy grenades ?
   if ((f_grenadebuy_time > 0) && (f_grenadebuy_time < gpGlobals->time))
   {
      // cycle through all clients and find bots
      for (client_index = 1; client_index <= gpGlobals->maxClients; client_index++)
      {
         pClient = INDEXENT (client_index);

         if (!pClient || pClient->free || !(pClient->v.flags & FL_FAKECLIENT))
            continue; // skip invalid clients and real players

         CLIENT_COMMAND (pClient, "buyitem\n"); // make the bot enter the buy item menu
         CLIENT_COMMAND (pClient, "menuselect 4\n"); // make the bot buy a grenade
      }

      f_grenadebuy_time = 0; // reset the timer so we won't do it twice
   }

   RETURN_META (MRES_IGNORED);
}


void RunPlayerMove (edict_t *fakeclient, const float *viewangles, float forwardmove, float sidemove, float upmove, unsigned short buttons, byte impulse, byte msec)
{
   char bot_weaponmodel[256];
   bool has_bomb, has_zoomable_rifle, has_scopable_rifle, has_grenade, has_knife;
   Vector v_potential_enemy = Vector (0, 0, 0);

   // retrieve the weapon the bot is holding in its hands
   strcpy (bot_weaponmodel, STRING (fakeclient->v.weaponmodel));

   // determine if it is holding a bomb, or any sort of scopable rifle, a grenade or a knife
   has_bomb = (strcmp (bot_weaponmodel, "models/p_c4.mdl") == 0);
   has_zoomable_rifle = (!has_bomb
                         && ((strcmp (bot_weaponmodel, "models/p_aug.mdl") == 0)
                             || (strcmp (bot_weaponmodel, "models/p_sg552.mdl") == 0)));
   has_scopable_rifle = (!has_bomb && !has_zoomable_rifle
                         && ((strcmp (bot_weaponmodel, "models/p_awp.mdl") == 0)
                             || (strcmp (bot_weaponmodel, "models/p_g3sg1.mdl") == 0)
                             || (strcmp (bot_weaponmodel, "models/p_scout.mdl") == 0)
                             || (strcmp (bot_weaponmodel, "models/p_sg550.mdl") == 0)));
   has_grenade = (!has_bomb && !has_zoomable_rifle && !has_scopable_rifle
                  && ((strcmp (bot_weaponmodel, "models/p_flashbang.mdl") == 0)
                      || (strcmp (bot_weaponmodel, "models/p_he.mdl") == 0)
                      || (strcmp (bot_weaponmodel, "models/p_smokegrenade.mdl") == 0)));
   has_knife = (!has_bomb && !has_zoomable_rifle && !has_scopable_rifle && !has_grenade
                && (strcmp (bot_weaponmodel, "models/p_knife.mdl") == 0));

   v_potential_enemy = FakeclientSeesEnemyPlayer (fakeclient);

   // is this bot about to plant the bomb ?
   if (has_bomb)
   {
      fakeclient->v.button = (fakeclient->v.button | IN_ATTACK) & ~IN_JUMP; // force it to plant the bomb right here right now
      (*g_engfuncs.pfnRunPlayerMove) (fakeclient, viewangles, forwardmove, sidemove, upmove, fakeclient->v.button, impulse, msec);
      RETURN_META (MRES_SUPERCEDE);
   }

   // else is this bot holding a sniper rifle AND we take over the sniping code ?
   if ((has_zoomable_rifle || has_scopable_rifle) && (use_scope > 0))
   {
      // is this bot trying to lock onto an enemy ?
      if (v_potential_enemy != Vector (0, 0, 0))
      {
         // is it a scoped weapon (sniper rifle) ?
         if (has_scopable_rifle)
         {
            // if the enemy is very far and bot is allowed to scope full zoom and bot is not scoping yet...
            if (((v_potential_enemy - fakeclient->v.origin).Length () > 600) && (use_scope > 1) && (fakeclient->v.fov > FOV_SCOPED_FAR))
            {
               if (RANDOM_LONG (0, 100) > 50)
                  fakeclient->v.button = (fakeclient->v.button | IN_ATTACK2) & ~IN_ATTACK; // scope until the highest zoom mode
               else
                  fakeclient->v.button &= IN_ATTACK2;
               (*g_engfuncs.pfnRunPlayerMove) (fakeclient, viewangles, forwardmove, sidemove, upmove, fakeclient->v.button, impulse, msec);
               RETURN_META (MRES_SUPERCEDE);
            }

            // else if it's not immediately nearby...
            else if (((v_potential_enemy - fakeclient->v.origin).Length () > 150) && (fakeclient->v.fov > FOV_SCOPED_SHORT))
            {
               if (RANDOM_LONG (0, 100) > 50)
                  fakeclient->v.button = (fakeclient->v.button | IN_ATTACK2) & ~IN_ATTACK; // scope to the first zoom mode (because scoped weapons have no crosshair, and bot NEEDS a crosshair :))
               else
                  fakeclient->v.button &= IN_ATTACK2;
               (*g_engfuncs.pfnRunPlayerMove) (fakeclient, viewangles, forwardmove, sidemove, upmove, fakeclient->v.button, impulse, msec);
               RETURN_META (MRES_SUPERCEDE);
            }
         }

         // else it is a zoomed weapon (Steyr Aug, Sig SG-552)
         else if (((v_potential_enemy - fakeclient->v.origin).Length () > 400) && (fakeclient->v.fov > FOV_ZOOMED))
         {
            if (RANDOM_LONG (0, 100) > 50)
               fakeclient->v.button = (fakeclient->v.button | IN_ATTACK2) & ~IN_ATTACK; // use the zoom when appropriate
            else
               fakeclient->v.button &= IN_ATTACK2;
            (*g_engfuncs.pfnRunPlayerMove) (fakeclient, viewangles, forwardmove, sidemove, upmove, fakeclient->v.button, impulse, msec);
            RETURN_META (MRES_SUPERCEDE);
         }
      }

      // else if the bot seems not to try to lock onto anyone, reset the weapon zoom to normal
      else if (fakeclient->v.fov < FOV_NORMAL)
      {
         if (RANDOM_LONG (0, 100) > 50)
            fakeclient->v.button = (fakeclient->v.button | IN_ATTACK2) & ~IN_ATTACK; // force it to reset the zoom right now so that the bot sees where it put its feet, at least!
         else
            fakeclient->v.button &= IN_ATTACK2;
         (*g_engfuncs.pfnRunPlayerMove) (fakeclient, viewangles, forwardmove, sidemove, upmove, fakeclient->v.button, impulse, msec);
         RETURN_META (MRES_SUPERCEDE);
      }

      // when everything is right, don't let the bot code handle the scope
      fakeclient->v.button &= ~IN_ATTACK2; // prevent POD-bot from zooming or unzooming on its own
      (*g_engfuncs.pfnRunPlayerMove) (fakeclient, viewangles, forwardmove, sidemove, upmove, fakeclient->v.button, impulse, msec);
      RETURN_META (MRES_SUPERCEDE);
   }

   // else if it's holding a grenade and the enemy is in sight, there's an emergency to throw it !
   else if (has_grenade && (v_potential_enemy != Vector (0, 0, 0)))
   {
      if (RANDOM_LONG (0, 100) > 50)
         fakeclient->v.button |= IN_ATTACK; // force it to throw the grenade right now
      else
         fakeclient->v.button &= ~IN_ATTACK;
      (*g_engfuncs.pfnRunPlayerMove) (fakeclient, viewangles, forwardmove, sidemove, upmove, fakeclient->v.button, impulse, msec);
      RETURN_META (MRES_SUPERCEDE);
   }

   // else if it's holding a knife, make it stab occasionally (more realistic ? also helps fixing stuck to enemy bug)
   else if (has_knife && (v_potential_enemy != Vector (0, 0, 0)))
   {
      if (RANDOM_LONG (0, 100) > 98)
         fakeclient->v.button |= IN_ATTACK; // force it to stab right now
      (*g_engfuncs.pfnRunPlayerMove) (fakeclient, viewangles, forwardmove, sidemove, upmove, fakeclient->v.button, impulse, msec);
      RETURN_META (MRES_SUPERCEDE);
   }

   // additionnally, if the bot is still in freeze time, it's no use jumping nor ducking :)
   if (f_roundstart_time > gpGlobals->time)
   {
      fakeclient->v.button &= ~(IN_JUMP | IN_DUCK); // prevent the bot from jumping like crazy
      (*g_engfuncs.pfnRunPlayerMove) (fakeclient, viewangles, forwardmove, sidemove, upmove, fakeclient->v.button, impulse, msec);
      RETURN_META (MRES_SUPERCEDE);
   }

   RETURN_META (MRES_IGNORED);
}


void TraceLine (const float *v1, const float *v2, int fNoMonsters, edict_t *pentToSkip, TraceResult *ptr)
{
   TRACE_LINE (v1, v2, fNoMonsters, pentToSkip, ptr);

   // check to see if a hostage was hit	
   if (strcmp (STRING (ptr->pHit->v.classname), "hostage_entity") == 0 )
   {
      ptr->flFraction = 0; // don't even reach the hostage, disallowing damage for it
      ptr->pHit->v.health = 0x0FFFFFFF; // max out the hostage life points again
      ptr->pHit->v.max_health = 0x0FFFFFFF;
      RETURN_META (MRES_SUPERCEDE);
   }

   RETURN_META (MRES_IGNORED);
}


void MessageBegin (int msg_dest, int msg_type, const float *pOrigin, edict_t *ed)
{
   // Bots don't have client g_args. It screws up my ClientCommand interception code, that's
   // why I have to catch these packets on the fly. Hence this awful hack (c) (r) (tm)...

   if (use_radio)
      RETURN_META (MRES_IGNORED); // no use catching messages if we still want bot radio

   // ANY message that passes through the podbot DLL is suspicious, that's what I say! :)

   suspicious_message_size = 0; // first set suspicious message size to zero

   // remember its parameters too...
   suspicious_message_header.msg_dest = msg_dest;
   suspicious_message_header.msg_type = msg_type;
   suspicious_message_header.pOrigin = pOrigin;
   suspicious_message_header.ed = ed;

   RETURN_META (MRES_SUPERCEDE); // but don't send it yet
}


void MessageEnd (void)
{
   if (use_radio)
      RETURN_META (MRES_IGNORED); // no use catching messages if we still want bot radio

   RETURN_META (MRES_SUPERCEDE); // don't send the packet
}


void MessageEnd_Post (void)
{
   if (use_radio)
      RETURN_META (MRES_IGNORED); // no use catching messages if we still want bot radio

   ExaminateSuspiciousMessage (); // now we've received the full message, check it before sending it

   RETURN_META (MRES_IGNORED);
}


void WriteByte (int iValue)
{
   if (use_radio)
      RETURN_META (MRES_IGNORED); // no use catching messages if we still want bot radio

   suspicious_message[suspicious_message_size].packet_type = PACKET_BYTE; // remember packet type
   suspicious_message[suspicious_message_size].iValuePassed = iValue; // remember value passed
   suspicious_message_size++; // the suspicious message has gained one packet in size

   RETURN_META (MRES_SUPERCEDE); // don't send the packet
}


void WriteChar (int iValue)
{
   if (use_radio)
      RETURN_META (MRES_IGNORED); // no use catching messages if we still want bot radio

   suspicious_message[suspicious_message_size].packet_type = PACKET_CHAR; // remember packet type
   suspicious_message[suspicious_message_size].iValuePassed = iValue; // remember value passed
   suspicious_message_size++; // the suspicious message has gained one packet in size

   RETURN_META (MRES_SUPERCEDE); // don't send the packet
}


void WriteShort (int iValue)
{
   if (use_radio)
      RETURN_META (MRES_IGNORED); // no use catching messages if we still want bot radio

   suspicious_message[suspicious_message_size].packet_type = PACKET_SHORT; // remember packet type
   suspicious_message[suspicious_message_size].iValuePassed = iValue; // remember value passed
   suspicious_message_size++; // the suspicious message has gained one packet in size

   RETURN_META (MRES_SUPERCEDE); // don't send the packet
}


void WriteLong (int iValue)
{
   if (use_radio)
      RETURN_META (MRES_IGNORED); // no use catching messages if we still want bot radio

   suspicious_message[suspicious_message_size].packet_type = PACKET_LONG; // remember packet type
   suspicious_message[suspicious_message_size].iValuePassed = iValue; // remember value passed
   suspicious_message_size++; // the suspicious message has gained one packet in size

   RETURN_META (MRES_SUPERCEDE); // don't send the packet
}


void WriteAngle (float flValue)
{
   if (use_radio)
      RETURN_META (MRES_IGNORED); // no use catching messages if we still want bot radio

   suspicious_message[suspicious_message_size].packet_type = PACKET_ANGLE; // remember packet type
   suspicious_message[suspicious_message_size].flValuePassed = flValue; // remember value passed
   suspicious_message_size++; // the suspicious message has gained one packet in size

   RETURN_META (MRES_SUPERCEDE); // don't send the packet
}


void WriteCoord (float flValue)
{
   if (use_radio)
      RETURN_META (MRES_IGNORED); // no use catching messages if we still want bot radio

   suspicious_message[suspicious_message_size].packet_type = PACKET_COORD; // remember packet type
   suspicious_message[suspicious_message_size].flValuePassed = flValue; // remember value passed
   suspicious_message_size++; // the suspicious message has gained one packet in size

   RETURN_META (MRES_SUPERCEDE); // don't send the packet
}


void WriteString (const char *sz)
{
   if (use_radio)
      RETURN_META (MRES_IGNORED); // no use catching messages if we still want bot radio

   suspicious_message[suspicious_message_size].packet_type = PACKET_STRING; // remember packet type
   suspicious_message[suspicious_message_size].szValuePassed = sz; // remember value passed
   suspicious_message_size++; // the suspicious message has gained one packet in size

   RETURN_META (MRES_SUPERCEDE); // don't send the packet
}


void WriteEntity (int iValue)
{
   if (use_radio)
      RETURN_META (MRES_IGNORED); // no use catching messages if we still want bot radio

   suspicious_message[suspicious_message_size].packet_type = PACKET_ENTITY; // remember packet type
   suspicious_message[suspicious_message_size].iValuePassed = iValue; // remember value passed
   suspicious_message_size++; // the suspicious message has gained one packet in size

   RETURN_META (MRES_SUPERCEDE); // don't send the packet
}


void ExaminateSuspiciousMessage (void)
{
   int index, client_index;
   edict_t *pPlayer = NULL;
   bool dont_send_message = FALSE;
   static bool dont_send_next_SendAudio_message = FALSE;
   static bool dont_send_next_TempEntity_message = FALSE;

   // is it sent to a valid client AND is it a RoundTime message ?
   if (!FNullEnt (suspicious_message_header.ed)
       && (suspicious_message_header.ed->v.flags & FL_CLIENT)
       && (suspicious_message_header.msg_type == GET_USER_MSG_ID (PLID, "RoundTime", NULL)))
   {
      // if we're resetting the map and we want the bots to buy more grenades...

      f_roundstart_time = gpGlobals->time + CVAR_GET_FLOAT ("mp_freezetime"); // compute round start date

      if (buy_grenades)
         f_grenadebuy_time = gpGlobals->time + 3.0; // make them buy a grenade in 4 seconds
      else
         f_grenadebuy_time = 0;
   }

   // if we don't want bots to use the radio, we have to check this packet
   if (!use_radio)
   {
      // is it a TextMsg message ?
      if (suspicious_message_header.msg_type == GET_USER_MSG_ID (PLID, "TextMsg", NULL))
      {
         // first check if it is a radio message
         for (index = 0; index < suspicious_message_size; index++)
         {
            // is it a radio message ?
            if ((suspicious_message[index].packet_type == PACKET_STRING)
                && (strcmp (suspicious_message[index].szValuePassed, "#Game_radio") == 0))
            {
               // now look if that radio message has a sender specified
               for (index; index < suspicious_message_size; index++)
               {
                  // next string in the message should be the sender name...
                  if (suspicious_message[index].packet_type == PACKET_STRING)
                  {
                     // ...so cycle through all players
                     for (client_index = 1; client_index <= gpGlobals->maxClients; client_index++)
                     {
                        pPlayer = INDEXENT (client_index);

                        if (FNullEnt (pPlayer) || pPlayer->free)
                           continue; // skip invalid players

                        // is this player the sender of the message AND this player is a bot ?
                        if ((strcmp (suspicious_message[index].szValuePassed, STRING (pPlayer->v.netname)) == 0) && (pPlayer->v.flags & FL_FAKECLIENT))
                        {
                           // then look if it's NOT a grenade throwing related message NOR a generic radio message
                           for (index = 0; index < suspicious_message_size; index++)
                              if ((suspicious_message[index].packet_type == PACKET_STRING)
                                  && (strcmp (suspicious_message[index].szValuePassed, "#Bomb_Planted") != 0)
                                  && (strcmp (suspicious_message[index].szValuePassed, "#Bomb_Defused") != 0)
                                  && (strcmp (suspicious_message[index].szValuePassed, "#Round_Draw") != 0)
                                  && (strcmp (suspicious_message[index].szValuePassed, "#Terrorists_Win") != 0)
                                  && (strcmp (suspicious_message[index].szValuePassed, "#CTs_Win") != 0)
                                  && (strcmp (suspicious_message[index].szValuePassed, "#All_Hostages_Rescued") != 0))
                              {
                                 dont_send_message = TRUE; // then don't send this message
                                 dont_send_next_SendAudio_message = TRUE; // nor the next SendAudio message
                                 dont_send_next_TempEntity_message = TRUE; // nor the next TempEntity message
                              }
                        }
                     }
                  }
               }
            }
         }
      }

      // else is it a SendAudio message ?
      else if (suspicious_message_header.msg_type == GET_USER_MSG_ID (PLID, "SendAudio", NULL))
      {
         // have we been warned it was a bot radio message ?
         if (dont_send_next_SendAudio_message)
         {
            dont_send_message = TRUE; // then block this message
            dont_send_next_SendAudio_message = FALSE; // don't forget to reset the warning flag
         }
      }

      // else is it a tempentity message ?
      else if (suspicious_message_header.msg_type == GET_USER_MSG_ID (PLID, "tempentity?", NULL))
      {
         // have we been warned it was a bot radio message ?
         if (dont_send_next_TempEntity_message)
         {
            dont_send_message = TRUE; // then block this message
            dont_send_next_TempEntity_message = FALSE; // don't forget to reset the warning flag
         }
      }
   }

   // now send the message if we're allowed to
   if (!dont_send_message)
   {
      MESSAGE_BEGIN (suspicious_message_header.msg_dest, suspicious_message_header.msg_type, suspicious_message_header.pOrigin, suspicious_message_header.ed);
      for (index = 0; index < suspicious_message_size; index++)
      {
         if (suspicious_message[index].packet_type == PACKET_BYTE)
            WRITE_BYTE (suspicious_message[index].iValuePassed);
         else if (suspicious_message[index].packet_type == PACKET_CHAR)
            WRITE_CHAR (suspicious_message[index].iValuePassed);
         else if (suspicious_message[index].packet_type == PACKET_SHORT)
            WRITE_SHORT (suspicious_message[index].iValuePassed);
         else if (suspicious_message[index].packet_type == PACKET_LONG)
            WRITE_LONG (suspicious_message[index].iValuePassed);
         else if (suspicious_message[index].packet_type == PACKET_ANGLE)
            WRITE_ANGLE (suspicious_message[index].flValuePassed);
         else if (suspicious_message[index].packet_type == PACKET_COORD)
            WRITE_COORD (suspicious_message[index].flValuePassed);
         else if (suspicious_message[index].packet_type == PACKET_STRING)
            WRITE_STRING (suspicious_message[index].szValuePassed);
         else if (suspicious_message[index].packet_type == PACKET_ENTITY)
            WRITE_ENTITY (suspicious_message[index].iValuePassed);
      }
      MESSAGE_END ();
   }

   return; // finished
}


Vector FakeclientSeesEnemyPlayer (edict_t *pFakeclient)
{
   int client_index;
   edict_t *pPlayer = NULL;
   TraceResult tr;

   if (FNullEnt (pFakeclient))
      return (Vector (0, 0, 0)); // reliability check

   // cycle through all players
   for (client_index = 1; client_index < gpGlobals->maxClients; client_index++)
   {
      pPlayer = INDEXENT (client_index);

      if (FNullEnt (pPlayer) || pPlayer->free || (pPlayer == pFakeclient)
          || (pPlayer->v.effects & EF_NODRAW) || (GetTeam (pFakeclient) == GetTeam (pPlayer)))
         continue; // skip invalid players, invisible ones, ourselves and teammates

      MAKE_VECTORS (pFakeclient->v.v_angle);
      if (UTIL_AngleOfVectors (gpGlobals->v_forward, pPlayer->v.origin - pFakeclient->v.origin) > pFakeclient->v.fov / 2)
         continue; // skip players outside bot's field of view

      MAKE_VECTORS (UTIL_VecToAngles (pPlayer->v.origin - pFakeclient->v.origin));

      // trace a line from bot's eye to this player's head
      TRACE_LINE (pFakeclient->v.origin + pFakeclient->v.view_ofs, pPlayer->v.origin + pPlayer->v.view_ofs, ignore_monsters, pFakeclient, &tr);
      if (tr.flFraction == 1.0)
         return (tr.vecEndPos); // enemy in sight if the line traced all the way up to its destination

      // trace a line from bot's eye to this player's body
      TRACE_LINE (pFakeclient->v.origin + pFakeclient->v.view_ofs, pPlayer->v.origin, ignore_monsters, pFakeclient, &tr);
      if (tr.flFraction == 1.0)
         return (tr.vecEndPos); // enemy in sight if the line traced all the way up to its destination

      // trace a line from bot's eye to this player's left arm
      TRACE_LINE (pFakeclient->v.origin + pFakeclient->v.view_ofs, pPlayer->v.origin - (gpGlobals->v_right * 16), ignore_monsters, pFakeclient, &tr);
      if (tr.flFraction == 1.0)
         return (tr.vecEndPos); // enemy in sight if the line traced all the way up to its destination

      // trace a line from bot's eye to this player's right arm
      TRACE_LINE (pFakeclient->v.origin + pFakeclient->v.view_ofs, pPlayer->v.origin + (gpGlobals->v_right * 16), ignore_monsters, pFakeclient, &tr);
      if (tr.flFraction == 1.0)
         return (tr.vecEndPos); // enemy in sight if the line traced all the way up to its destination
   }

   return (Vector (0, 0, 0)); // looks like this bot has no enemy in sight
}


int GetTeam (edict_t *pEntity)
{
   char model_name[32];
   strcpy (model_name, INFOKEY_VALUE (GET_INFOKEYBUFFER (pEntity), "model"));

   if ((strcmp (model_name, "terror") == 0) // Phoenix Connection
       || (strcmp (model_name, "leet") == 0) // L337 Krew
       || (strcmp (model_name, "arctic") == 0) // Arctic Avenger
       || (strcmp (model_name, "guerilla") == 0)) // Guerilla Warfare
      return (TEAM_TERRORIST);
   else
      return (TEAM_COUNTER_TERRORIST);
}


Vector UTIL_VecToAngles (const Vector &vec)
{
   float rgflVecOut[3];

   // avoid null vectors
   if (vec != Vector (0, 0, 0))
      VEC_TO_ANGLES (vec, rgflVecOut);

   return (Vector (rgflVecOut));
}


float UTIL_AngleOfVectors (Vector v1, Vector v2)
{
   if ((v1.Length () == 0) || (v2.Length () == 0))
      return 0; // avoid zero divide

   return (abs (UTIL_WrapAngle (acos (DotProduct (v1, v2) / (v1.Length () * v2.Length ())) * 180 / M_PI)));
}


float UTIL_WrapAngle (float angle_to_wrap)
{
   float angle = angle_to_wrap;

   // check for wraparound of angle
   if (angle >= 180)
      angle -= 360 * ((int) (angle / 360) + 1);
   if (angle < -180)
      angle += 360 * ((int) (angle / 360) + 1);

   return angle;
}


void GetGameDir (char *game_dir)
{
   // This function fixes the erratic behaviour caused by the use of the GET_GAME_DIR engine
   // macro, which returns either an absolute directory path, or a relative one, depending on
   // whether the game server is run standalone or not. This one always return a RELATIVE path.

   unsigned char length, fieldstart, fieldstop;

   GET_GAME_DIR (game_dir); // call the engine macro and let it mallocate for the char pointer

   length = strlen (game_dir); // get the length of the returned string
   length--; // ignore the trailing string terminator

   // format the returned string to get the last directory name
   fieldstop = length;
   while (((game_dir[fieldstop] == '\\') || (game_dir[fieldstop] == '/')) && (fieldstop > 0))
      fieldstop--; // shift back any trailing separator

   fieldstart = fieldstop;
   while ((game_dir[fieldstart] != '\\') && (game_dir[fieldstart] != '/') && (fieldstart > 0))
      fieldstart--; // shift back to the start of the last subdirectory name

   if ((game_dir[fieldstart] == '\\') || (game_dir[fieldstart] == '/'))
      fieldstart++; // if we reached a separator, step over it

   // now copy the formatted string back onto itself character per character
   for (length = fieldstart; length <= fieldstop; length++)
      game_dir[length - fieldstart] = game_dir[length];
   game_dir[length - fieldstart] = 0; // terminate the string

   return;
}
